"""
Docstring it
"""
import calendar
import logging
import socket
import struct
import time

from .bal_crc import CRC


# pylint: disable=too-many-instance-attributes
# 29 is reasonable in this case.
class Packet:
    """
    Packet class
    """
    header_struct = struct.Struct('!LBBHHH')
    assert header_struct.size == 12

    brdy_cmd_code = 0x5A
    brdy_struct = struct.Struct('!LL2s5sBHHLQ')
    assert brdy_struct.size == 32

    vack_struct = struct.Struct('!5L6HQ')
    assert vack_struct.size == 40

    baler_response_code = 0x5C
    baler_response_struct = struct.Struct('!HHL')
    assert baler_response_struct.size == 8

    seq_counter = 666

    def __init__(self):
        self.access_timeout = None
        self.ack_number = None
        self.baler_command = None
        self.baler_ip = None
        self.baler_netmask = None
        self.baler_sn = None
        self.baler_version = None
        self.bps = None
        self.buf = None
        self.command = None
        self.crc = None
        self.disk_size = None
        self.flags = None
        self.length = None
        self.model_num = None
        self.net = None
        self.payload = None
        self.proto_version = None
        self.q330_base_port = None
        self.q330_data_port = None
        self.q330_ip = None
        self.q330_sn1 = None
        self.q330_sn2 = None
        self.seq_number = None
        self.sequencing_field = None
        self.serial_baud_rate = None
        self.station_name = None
        self.time = None

    def crc_check(self):
        """
        CRC check.
        """
        if self.crc == CRC.compute_crc(self.buf[4:]):
            logging.debug('crc: pass')
            return True
        logging.warning('crc: fail')
        return False

    def unpack(self, buf):
        """
        Docstring
        """
        self.buf = buf
        self.unpack_header(buf[:self.header_struct.size])
        self.crc_check()
        if self.command == self.brdy_cmd_code:
            self.unpack_brdy(buf[self.header_struct.size:])

    def unpack_header(self, buf):
        """
        Docstring
        """
        (self.crc,
         self.command,
         self.proto_version,
         self.length,
         self.seq_number,
         self.ack_number) = self.header_struct.unpack(buf)

    def unpack_brdy(self, buf):
        """
        Docstring this function
        """
        (self.q330_sn1,
         self.q330_sn2,
         self.net,
         self.station_name,
         self.flags,
         self.model_num,
         self.baler_version,
         self.disk_size,
         self.baler_sn) = self.brdy_struct.unpack(buf)

    def pack_vack(self, ip_addr, netmask, brdy_pkt):
        """
        Docstring this function
        """
        assert isinstance(brdy_pkt, self.__class__)
        self.q330_sn1 = 0
        self.q330_sn2 = brdy_pkt.q330_sn2
        self.q330_ip = 0
        self.baler_ip = ip_addr
        self.baler_netmask = str(netmask)
        self.q330_base_port = 0
        self.q330_data_port = 0
        self.bps = 0
        self.flags = 0
        self.access_timeout = 0
        self.serial_baud_rate = 0
        self.baler_sn = brdy_pkt.baler_sn
        self.payload = self.vack_struct.pack(
            self.q330_sn1,
            self.q330_sn2,
            self.q330_ip,
            struct.unpack('!I', socket.inet_aton(self.baler_ip))[0],
            struct.unpack('!I', socket.inet_aton(self.baler_netmask))[0],
            self.q330_base_port,
            self.q330_data_port,
            self.bps,
            self.flags,
            self.access_timeout,
            self.serial_baud_rate,
            self.baler_sn)
        # header
        self.command = 0xC7
        self.proto_version = brdy_pkt.proto_version
        self.ack_number = brdy_pkt.seq_number
        self.pack_header()

    def pack_ping(self):
        """
        An empty command packet useful to ping baler
        """
        self.baler_command = 0
        self.sequencing_field = 0
        self.flags = 0
        clean_struct = struct.Struct('!2HL')
        self.payload = clean_struct.pack(self.baler_command,
                                         self.sequencing_field,
                                         self.flags)
        # header C8=baler command.
        self.command = 0xC8
        # XXX NOTE: assume v2
        self.proto_version = 2
        self.ack_number = 0
        self.pack_header()

    def pack_set_baler_time(self, brdy_pkt):
        """
        Docstring this function
        """
        base_time = calendar.timegm((2000, 1, 1, 0, 0, 0, 0, 0, 0))
        now = time.time() - base_time
        self.baler_command = 5
        self.sequencing_field = 0  # used for multiple packet cmds
        self.time = int(now)
        # XXX TODO: move to class
        time_struct = struct.Struct('!2HL')
        self.payload = time_struct.pack(self.baler_command,
                                        self.sequencing_field,
                                        self.time)
        # header
        self.command = 0xC8
        self.proto_version = brdy_pkt.proto_version
        self.ack_number = 0
        self.pack_header()

    def pack_shutdown(self):
        """
        Turn baler off qlib calls deallocate
        """
        self.baler_command = 4
        self.sequencing_field = 0
        self.flags = 0
        # TODO: Test to see the differences
        clean_struct = struct.Struct('!H')
        self.payload = clean_struct.pack(self.baler_command)
        # clean_struct = struct.Struct('!2HL')
        # self.payload = clean_struct.pack(self.baler_command,
        #                                  self.sequencing_field,
        #                                  self.flags)
        # header
        self.command = 0xC8
        # XXX NOTE: assume v2
        self.proto_version = 2
        self.ack_number = 0
        self.pack_header()

    def pack_clean(self):
        """
        Clean the baler and remove q330 association
        """
        self.baler_command = 2
        self.flags = 0xff
        self.sequencing_field = 0
        # TODO: Test to see the differences
        # clean_struct = struct.Struct('!2HL')
        clean_struct = struct.Struct('!3H')
        self.payload = clean_struct.pack(self.baler_command,
                                         self.sequencing_field,
                                         self.flags)
        # header
        self.command = 0xC8
        # XXX NOTE: assume v2
        self.proto_version = 2
        self.ack_number = 0
        self.pack_header()

    def pack_header(self):
        """
        Docstring this.
        """
        self.crc = 0
        self.length = len(self.payload)
        self.seq_counter += 1
        header_bytes = self.header_struct.pack(
            self.crc,
            self.command,
            self.proto_version,
            self.length,
            self.seq_counter,
            self.ack_number)
        qdp_packet = header_bytes + self.payload
        # compute and insert crc
        to_crc = qdp_packet[4:]
        self.crc = CRC.compute_crc(to_crc)
        self.buf = struct.pack('!L', self.crc) + to_crc

    def __str__(self):
        ret = 'QDP packet:'
        for key, value in self.__dict__.items():
            if key in ('buf', 'payload'):
                continue
            ret += f'\t{key:20}:\t{value}\n'
        return ret
