"""
This module provides general utilities for the bline project.
Author: Omid Hosseini
Started Date: 02/26/2021
"""
import errno
import logging
import os
import sys
import re
import bs4
import requests
from bs4 import BeautifulSoup

# Grab the logger.
# The logger needs to be configured correctly.
# TODO: CHANGE THE CONFIGS TO CAPTURE THE RIGHT LOGS
# TODO: ADD LOGGING TO DIFFERENT METHODS.
logger = logging.getLogger(__name__)


def retrieve_baler_dir_list(ip_addr):
    """
    Return the Baler's Directory listing htm file content.

    keyword arguments:
    ip_addr -- (str) the specific ip address to search for the baler info.
    """
    dir_list = dict()
    # TODO: FIND THE RIGHT ADDRESS FOR THE DIRECTORY LISTING HTM FILE
    soup = get_htm(ip_addr, "****.htm")
    # Grab the "pre" tag in the htm file. The ham file is designed badly and
    # there is overflow all the time. The function works fine.
    pre_text = soup.find("pre")
    # Loop through the elements in the pre_text. All odd indices are tags and
    # all even indices are navigableStrings
    for line in pre_text:
        if isinstance(line, bs4.element.Tag):
            # The structure of the tags are like:
            # <a href="http://192.168.67.37/DT0001__.BHZ">DT0001__.BHZ</a>
            # Tag's text is used as the dict key and the url is value.
            # size, from, and to will be obtained in the next iteration.
            dic_key = str(line.text)
            dir_list[dic_key] = {"address": line['href'], "size": "",
                                 "from": "", "to": ""}
        if isinstance(line, bs4.element.NavigableString):
            # The structre of the navigableString is like:
            #   16777216 bytes, from 2021-03-08 21:20:28 to 2021-03-12 00:36:31
            # The following regex split will clean the string into three pieces
            # [size, from, to]
            extracted_text = \
                re.split(
                    '^\s+|\s*bytes\s*|\s*,\s*|\s*from\s*|\s*to\s*|\s*\n\s*',
                    line.string)
            # Get rid of the empty split elements
            extracted_text = [x for x in extracted_text if x]
            # Complete the dictionary's entry.
            dir_list[dic_key]["size"] = extracted_text[0]
            dir_list[dic_key]["from"] = extracted_text[1]
            dir_list[dic_key]["to"] = extracted_text[2]
    return dir_list


def clear_baler_status(ip_addr):
    """
    Clear the Baler's interface status.

    keyword arguments:
    ip_addr -- (str) the specific ip address to search for the baler info.
    """
    # TODO: IMPLEMENTATION IS NEEDED. CHECK BALER'S INTERFACE STATUS HTM FILE.
    pass


def retrieve_baler_thrd_status(ip_addr):
    """
    Return the Baler's thread status.

    keyword arguments:
    ip_addr -- (str) the specific ip address to search for the baler info.
    """
    baler_thrd_status = list()
    attribute_names = list()

    # TODO: FIND THE RIGHT ADDRESS FOR THE BALER THREAD STATUS HTM FILE
    soup = get_htm(ip_addr, "****.htm")
    # Grab the table title
    table_title = soup.find("pre").contents[0].text
    # Grab the table body
    # Really bad htm design and created a lot of headache to grab the right
    # attribute names and rows.
    # To understand this you should check the htm file.
    table_body = soup.find("pre").contents[1]
    # First split the attribute names based on the white space and then attach
    # the first two element to create 9 attribute names.
    # The format will be like:
    # Thread Name|Blk|Pri|Cnt|Evt|Last-Run|Intrpt|Percent|Total-Time
    tmp = table_title.split()
    attribute_names.append(tmp[0] + " " + tmp[1])
    attribute_names.extend(tmp[2:10])
    baler_thrd_status.append(attribute_names)
    # the best way to split the rows are based on "\n". Four different cases
    # must be handled.
    rows = table_body.split("\n")
    # Check the structure of the htm file to understand the logic of this loop.
    for i in range(len(rows)):
        if i in {0, 12}:
            continue
        if i in {4, 7, 8, 9, 10}:
            baler_thrd_status.append(rows[i].split())
        elif i == 11:
            baler_thrd_status.append(rows[i])
        else:
            tmp = list()
            tmp.append(rows[i].split()[0] + " " + rows[i].split()[1])
            tmp.extend(rows[i].split()[2:10])
            baler_thrd_status.append(tmp)
    return baler_thrd_status


def retrieve_baler_status(ip_addr):
    """
    Retrieve the baler interface status from the htm file.

    keyword arguments:
    ip_addr -- (str) the specific ip address to search for the baler info.
    """
    # TODO: FIND THE RIGHT ADDRESS FOR THE BALER INTERFACE STATUS HTM FILE
    soup = get_htm(ip_addr, "****.htm")
    # Returns to dictionaries. Baler's ethernet status and serial connection
    # status.
    baler_eth_info = dict()
    baler_ser_info = dict()
    # Obtain the required fields.
    b_tag = soup.findAll('b')
    all_br_tags = b_tag[0].text.strip()
    text_list = all_br_tags.split('\n')

    for line in range(len(text_list)):
        # Line 0 - 17 are information about ethernet status
        if line >= 0 and line < 18:
            if line == 0:
                continue
            k, v = text_list[line].split(': ', 1)
            baler_eth_info[k] = v
        # Lines 18 - 27 are serial status
        if line >= 18:
            if line == 18:
                continue
            k, v = text_list[line].split(': ', 1)
            baler_ser_info[k] = v
    return baler_eth_info, baler_ser_info


def retrieve_baler_info(ip_addr):
    """
    Retrieve the baler General information from the baler.htm file.

    keyword arguments:
    ip_addr -- (str) the specific ip address to search for the baler info.
    """
    baler_info = {}
    soup = get_htm(ip_addr, "baler.htm")
    # Check the structure of the htm file.
    for info in soup.findAll('li'):
        # Inconsistent line format needs different types of splitting.
        # A line like:"DMU2 Software Version=4, Serial Number=0100000B65071A28"
        if len(find_occur(info.text, ',')) == 1 and len(find_occur(info.text,
                                                                   '=')) == 2:
            spl_str = info.text.split(', ')
            for i in spl_str:
                k, v = i.split('=')
                baler_info[k] = v
        # A line like: "Supply Voltage=12.8"
        elif len(find_occur(info.text, ',')) == 0 \
            and len(find_occur(info.text, '=')) == 1:
            k, v = info.text.split('=')
            baler_info[k] = v
        # A line like: "MAC Address: F2:3F:0B:65:07:1A"
        elif len(find_occur(info.text, ':')) > 0:
            k, v = info.text.split(': ', 1)
            baler_info[k] = v
    return baler_info


def find_occur(input_str, charc):
    """
    Return a list of the indices of str in which the charc occur.

    keyword arguments:
    strg -- (str) String to perform the search process
    charc -- (str) The substring to be found in the string.
    """
    return [i for i, letter in enumerate(input_str) if letter == charc]


def get_htm(ip_addr, page="baler.htm"):
    """
    Return a data structure representing a parsed HTML or XML document.

    keyword arguments:
    ip_addr -- (str) The IP address of the Baler
    page -- (str)
    """
    htm_address = "http://{}/{}}".format(str(ip_addr), str(page))
    htm_content = requests.get(htm_address).text
    soup = BeautifulSoup(htm_content, 'lxml')
    return soup


def check_tags(tag_id):
    """
    Check the format of user input for tag_ids

    keyword arguments:
    tag_id -- (tuple) List of user input for tag_ids
    """
    try:
        tag_id = tuple(int(tgs) for tgs in tag_id)
        for tgs in tag_id:
            if tgs < 0:
                logger.error("The input tag value: %d is non-positive!", tgs)
                return False
    except ValueError as err:
        logger.error("The input tags has invalid type. %s", err)
        return False
    return True


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


def user_decision_tags(missing_tags, extra_tags):
    """
    Process user input about the Baler tags.

    keywords arguments:
    missing_tags -- (set) List of Baler tag(s) that the user provided as an
    input but not identified by the bline.

    extra_tags -- (set) List of Baler tag(s) identified by the bline but the
    user didn't provide them as an input.
    """

    # Keep the user decisions in a dictionary and return it to the caller.
    # "conn_miss" --> Ask bline to connect the missing tag(s) if True,
    # otherwise ignore.
    # "kpt_extra" --> Ask bline to keep the extra found tag(s),
    # DON'T DISCONNECT THEM, if True, otherwise send out shutdown command to
    # them.
    usr_decision = {"conn_miss": False, "kpt_extra": False}
    # The user has entered tagIDs that don't match with the identified
    # ones. Report them and ask for the proper action.
    if len(missing_tags) != 0:
        logger.warning("Devices with the following tag(s), %s, not identified",
                       missing_tags)
        print("Baler(s) with following tag(s), %s, have not been "
              "addressed. (Not connected perhaps!)." % missing_tags)
        if query_yes_no("Would you like to connect the Baler(s) with the "
                        "missing tag(s) [Yes/No]?"):
            usr_decision["conn_miss"] = True
    # Extra devices with not listed tags are addressed by the bline.
    # report them and ask for the proper action.
    if len(extra_tags) != 0:
        logger.warning("Extra devices with tag(s), %s, are identified and "
                       "addressed.", extra_tags)
        print("Baler(s) with following tag(s), %s, are found but not "
              "list by the user (Device(s) with wrong tag(s) are "
              "connected.)." % extra_tags)
        if query_yes_no("would you like to keep connected Baler(s) with tag(s)"
                        " that are not listed in the user input [Yes/No]?"):
            usr_decision["kpt_extra"] = True
    return usr_decision


def mkdir_p(path):
    """
    Create a folder in a specific path.

    keyword arguments:
    path -- (str) path to directory the user wants to create.
    """
    try:
        os.makedirs(path, exist_ok=True)  # Python > 3.2
    except TypeError:
        try:
            os.makedirs(path)
        # Python > 2.5
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise
